<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Loader;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;

// Import SPL stuff
use \InvalidArgumentException;
use \RecursiveDirectoryIterator;
use \RecursiveIteratorIterator;
use \SplFileInfo;

/**
 * This class loads class include files with a specific prefix and suffix
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		1.6.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * ----------------------------------
 * 1.6.0
 *  - This class loader is now 100% singleton, no other instance is really
 *    required, therefore the factory method can be removed safely
 *  - renamed initLoader() to initClassLoader()
 *  - An instance of a FrameworkConfiguration is no longer set here as this
 *    violated the rule that there shall be no instance set of that class
 *  - .htaccess is marked as deprecated as this should be no longer done
 *  - scanClassPath() is now protected, please use other scan*() methods
 * 1.5.0
 *  - Namespace scheme Project\Package[\SubPackage...] is fully supported and
 *    throws an InvalidArgumentException if not present. The last part will be
 *    always the class' name.
 * 1.4.0
 *  - Some comments improved, other minor improvements
 * 1.3.0
 *  - Constructor is now empty and factory method 'createClassLoader' is created
 *  - renamed loadClasses to scanClassPath
 *  - Added initLoader()
 * 1.2.0
 *  - ClassLoader rewritten to PHP SPL's own RecursiveIteratorIterator class
 * 1.1.0
 *  - loadClasses rewritten to fix some notices
 * 1.0.0
 *  - Initial release
 * ----------------------------------
 */
final class ClassLoader {
	/**
	 * Instance of this class
	 */
	private static $selfInstance = NULL;

	/**
	 * Cached configuration entry 'developer_mode_enabled'
	 */
	private static $developerModeEnabled = false;

	/**
	 * Array with all valid but pending for loading file names (class,
	 * interfaces, traits must start with below prefix).
	 */
	private $pendingFiles = [];

	/**
	 * List of loaded classes
	 */
	private $loadedClasses = [];

	/**
	 * Suffix with extension for all class files
	 */
	private $prefix = 'class_';

	/**
	 * Suffix with extension for all class files
	 */
	private $suffix = '.php';

	/**
	 * A list for directory names (no leading/trailing slashes!) which not be scanned by the path scanner
	 * @see	scanLocalPath
	 */
	private $ignoreList = [];

	/**
	 * Debug this class loader? (true = yes, false = no)
	 */
	private $debug = false;

	/**
	 * Whether the file list is cached
	 */
	private $listCached = false;

	/**
	 * Wethe class content has been cached
	 */
	private $classesCached = false;

	/**
	 * SplFileInfo for the list cache
	 */
	private $listCacheFile = NULL;

	/**
	 * SplFileInfo for class content
	 */
	private $classCacheFile = NULL;

	/**
	 * Counter for loaded include files
	 */
	private $total = 0;

	/**
	 * By default the class loader is strict with naming-convention check
	 */
	private static $strictNamingConvention = true;

	/**
	 * Framework/application paths for classes, etc.
	 */
	private static $frameworkPaths = [
		'classes',    // Classes
		'exceptions', // Exceptions
		'interfaces', // Interfaces
		'middleware', // The middleware
		'traits',     // Traits
	];

	/**
	 * Registered paths where test classes can be found. These are all relative
	 * to base_path .
	 */
	private static $testPaths = [];

	/**
	 * The protected constructor. Please use the factory method below, or use
	 * getSelfInstance() for singleton
	 *
	 * @return	void
	 */
	private function __construct () {
		// Cache config entry
		self::$developerModeEnabled = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('developer_mode_enabled');
	}

	/**
	 * The destructor makes it sure all caches got flushed
	 *
	 * @return	void
	 */
	public function __destruct () {
		// Skip here if dev-mode
		if (self::$developerModeEnabled) {
			return;
		}

		// Skip here if already cached
		if ($this->listCached === false) {
			// Writes the cache file of our list away
			$cacheContent = json_encode($this->pendingFiles);

			// Open cache instance
			$fileObject = $this->listCacheFile->openFile('w');

			// And write whole list
			$fileObject->fwrite($cacheContent);
		}

		// Skip here if already cached
		if ($this->classesCached === false) {
			// Generate a full-cache of all classes
			$cacheContent = '';
			foreach (array_keys($this->loadedClasses) as $fileInstance) {
				// Open file
				$fileObject = $fileInstance->openFile('r');

				// Load the file
				// @TODO Add some uglifying code (compress) here
				$cacheContent .= $fileObject->fread($fileInstance->getSize());
			}

			// Open file
			$fileObject = $this->classCacheFile->openFile('w');

			// And write it away
			$fileObject->fwrite($cacheContent);
		}
	}

	/**
	 * Scans for all framework classes, exceptions and interfaces.
	 *
	 * @return	void
	 */
	public static function scanFrameworkClasses () {
		// Get loader instance
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$loaderInstance = self::getSelfInstance();

		// "Cache" configuration instance and framework base path
		$configInstance = FrameworkBootstrap::getConfigurationInstance();
		$frameworkBasePath = $configInstance->getConfigEntry('framework_base_path');

		// Load all classes
		//* NOISY-DEBUG: */ printf('[%s:%d]: frameworkBasePath=%s,self::$frameworkPaths()=%d,' . PHP_EOL, __METHOD__, __LINE__, $frameworkBasePath, count(self::$frameworkPaths));
		foreach (self::$frameworkPaths as $shortPath) {
			// Generate full path from it
			//* NOISY-DEBUG: */ printf('[%s:%d]: shortPath[%s]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype($shortPath), $shortPath);
			$realPathName = realpath(sprintf(
				'%smain%s%s%s',
				$frameworkBasePath,
				DIRECTORY_SEPARATOR,
				$shortPath,
				DIRECTORY_SEPARATOR
			));

			// Is it not false and accessible?
			//* NOISY-DEBUG: */ printf('[%s:%d]: realPathName[%s]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype($realPathName), $realPathName);
			if (is_bool($realPathName)) {
				// Skip this, it is not accessible
				continue;
			} elseif (!is_readable($realPathName)) {
				// @TODO Throw exception instead of break
				break;
			}

			// Try to load the framework classes
			$loaderInstance->scanClassPath($realPathName);
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Scans for application's classes, etc.
	 *
	 * @return	void
	 */
	public static function scanApplicationClasses () {
		// Get loader instance
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$loaderInstance = self::getSelfInstance();

		// "Cache" configuration instance
		$configInstance = FrameworkBootstrap::getConfigurationInstance();

		// Load all classes for the application
		foreach (self::$frameworkPaths as $shortPath) {
			// Create path name
			//* NOISY-DEBUG: */ printf('[%s:%d]: shortPath=%s' . PHP_EOL, __METHOD__, __LINE__, $shortPath);
			$pathName = realpath(sprintf(
				'%s%s%s%s%s',
				$configInstance->getConfigEntry('application_base_path'),
				DIRECTORY_SEPARATOR,
				FrameworkBootstrap::getDetectedApplicationName(),
				DIRECTORY_SEPARATOR,
				$shortPath
			));

			// Is the path readable?
			//* NOISY-DEBUG: */ printf('[%s:%d]: pathName[%s]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype($pathName), $pathName);
			if (is_dir($pathName)) {
				// Try to load the application classes
				$loaderInstance->scanClassPath($pathName);
			}
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Scans for test classes, etc.
	 *
	 * @return	void
	 */
	public static function scanTestsClasses () {
		// "Cache" configuration instance
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$configInstance = FrameworkBootstrap::getConfigurationInstance();

		// Load all classes for the application
		foreach (self::$testPaths as $shortPath) {
			// Construct path name
			//* NOISY-DEBUG: */ printf('[%s:%d]: shortPath=%s' . PHP_EOL, __METHOD__, __LINE__, $shortPath);
			$pathName = sprintf(
				'%s%s%s',
				$configInstance->getConfigEntry('root_base_path'),
				DIRECTORY_SEPARATOR,
				$shortPath
			);

			// Try to find it
			//* NOISY-DEBUG: */ printf('[%s:%d]: pathName[%s]=%s - BEFORE!' . PHP_EOL, __METHOD__, __LINE__, gettype($pathName), $pathName);
			$realPathName = realpath($pathName);

			// Is the path readable?
			//* NOISY-DEBUG: */ printf('[%s:%d]: realPathName[%s]=%s - AFTER!' . PHP_EOL, __METHOD__, __LINE__, gettype($realPathName), $realPathName);
			if ((is_dir($realPathName)) && (is_readable($realPathName))) {
				// Try to load the application classes
				ClassLoader::getSelfInstance()->scanClassPath($realPathName);
			}
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Enables or disables strict naming-convention tests on class loading
	 *
	 * @param	$strictNamingConvention	Whether to strictly check naming-convention
	 * @return	void
	 */
	public static function enableStrictNamingConventionCheck (bool $strictNamingConvention = true) {
		self::$strictNamingConvention = $strictNamingConvention;
	}

	/**
	 * Registeres given relative path where test classes reside. For regular
	 * framework uses, they should not be loaded (and used).
	 *
	 * @param	$relativePath	Relative path to test classes
	 * @return	void
	 */
	public static function registerTestsPath (string $relativePath) {
		// "Register" it
		//* NOISY-DEBUG: */ printf('[%s:%d]: relativePath=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $relativePath);
		self::$testPaths[$relativePath] = $relativePath;

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Autoload-function
	 *
	 * @param	$className	Name of the class to load
	 * @return	void
	 * @throws	InvalidArgumentException	If the class' name does not contain a namespace: Tld\Domain\Project is AT LEAST recommended!
	 */
	public static function autoLoad (string $className) {
		// Validate parameter
		//* NOISY-DEBUG: */ printf('[%s:%d] className=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $className);
		if (empty($className)) {
			// Should not be empty
			throw new InvalidArgumentException('Parameter "className" is empty');
		}

		// The class name MUST be at least Tld\Domain\Project\Package\SomeFooBar so split at them
		$classNameParts = explode("\\", $className);

		// At least 3 parts should be there
		if ((self::$strictNamingConvention === true) && (count($classNameParts) < 5)) {
			// Namespace scheme is: Tld\Domain\Project\Package[\SubPackage...]
			throw new InvalidArgumentException(sprintf('Class name "%s" is not conform to naming-convention: Tld\Domain\Project\Package[\SubPackage...]\SomeFooBar', $className));
		}

		// Try to include this class
		//* NOISY-DEBUG: */ printf('[%s:%d]: Calling self->loadClassFile(%s) ...' . PHP_EOL, __METHOD__, __LINE__, $className);
		self::getSelfInstance()->loadClassFile($className);

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Singleton getter for an instance of this class
	 *
	 * @return	$selfInstance	A singleton instance of this class
	 */
	public static final function getSelfInstance () {
		// Is the instance there?
		if (is_null(self::$selfInstance)) {
			// Get a new one and initialize it
			self::$selfInstance = new ClassLoader();
			self::$selfInstance->initClassLoader();
		}

		// Return the instance
		return self::$selfInstance;
	}

	/**
	 * Getter for total include counter
	 *
	 * @return	$total	Total loaded include files
	 */
	public final function getTotal () {
		return $this->total;
	}

	/**
	 * Getter for a printable list of included main/interfaces/exceptions
	 *
	 * @param	$includeList	A printable include list
	 */
	public function getPrintableIncludeList () {
		// Prepare the list
		$includeList = '';
		foreach ($this->loadedClasses as $classFile) {
			$includeList .= basename($classFile) . '<br />' . PHP_EOL;
		}

		// And return it
		return $includeList;
	}

	/**
	 * Scans recursively a local path for class files which must have a prefix and a suffix as given by $this->suffix and $this->prefix
	 *
	 * @param	$basePath		The relative base path to 'framework_base_path' constant for all classes
	 * @param	$ignoreList		An optional list (array forced) of directory and file names which shall be ignored
	 * @return	void
	 */
	protected function scanClassPath (string $basePath, array $ignoreList = [] ) {
		// Is a list has been restored from cache, don't read it again
		//* NOISY-DEBUG: */ printf('[%s:%d] basePath=%s,ignoreList()=%d - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $basePath, count($ignoreList));
		if ($this->listCached === true) {
			// Abort here
			//* NOISY-DEBUG: */ printf('[%s:%d] this->listCache=true - EXIT!' . PHP_EOL, __METHOD__, __LINE__);
			return;
		}

		// Keep it in class for later usage, but flip index<->value
		$this->ignoreList = array_flip($ignoreList);

		/*
		 * Set base directory which holds all our classes, an absolute path
		 * should be used here so is_dir(), is_file() and so on will always
		 * find the correct files and dirs.
		 */
		$basePath2 = realpath($basePath);

		// If the basePath is false it is invalid
		//* NOISY-DEBUG: */ printf('[%s:%d] basePath2[%s]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype($basePath2), $basePath2);
		if ($basePath2 === false) {
			/* @TODO: Do not exit here. */
			exit(__METHOD__ . ': Cannot read ' . $basePath . ' !' . PHP_EOL);
		} else {
			// Set base path
			$basePath = $basePath2;
		}

		// Get a new iterator
		//* NOISY-DEBUG: */ printf('[%s:%d] basePath=%s' . PHP_EOL, __METHOD__, __LINE__, $basePath);
		$iteratorInstance = new RecursiveIteratorIterator(new RecursiveDirectoryIterator($basePath), RecursiveIteratorIterator::CHILD_FIRST);

		// Load all entries
		while ($iteratorInstance->valid()) {
			// Get current entry
			$currentEntry = $iteratorInstance->current();

			// Get filename from iterator which is the class' name (according naming-convention)
			$fileName = $currentEntry->getFilename();

			// Current entry must be a file, not smaller than 100 bytes and not on ignore list
			if (!$currentEntry->isFile() || isset($this->ignoreList[$fileName]) || $currentEntry->getSize() < 100) {
				// Advance to next entry
				$iteratorInstance->next();

				// Skip non-file entries
				//* NOISY-DEBUG: */ printf('[%s:%d] SKIP: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
				continue;
			}

			// Is this file wanted?
			//* NOISY-DEBUG: */ printf('[%s:%d] FOUND: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
			if ((substr($fileName, 0, strlen($this->prefix)) == $this->prefix) && (substr($fileName, -strlen($this->suffix), strlen($this->suffix)) == $this->suffix)) {
				// Add it to the list
				//* NOISY-DEBUG: */ printf('[%s:%d] ADD: fileName=%s,currentEntry=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName, $currentEntry);
				$this->pendingFiles[$fileName] = $currentEntry;
			} else {
				// Not added
				//* NOISY-DEBUG: */ printf('[%s:%d] NOT ADDED: fileName=%s,currentEntry=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName, $currentEntry);
			}

			// Advance to next entry
			//* NOISY-DEBUG: */ printf('[%s:%d] NEXT: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
			$iteratorInstance->next();
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Initializes our loader class
	 *
	 * @return	void
	 */
	private function initClassLoader () {
		// Construct the FQFN for the cache
		if (!self::$developerModeEnabled) {
			// Init cache instances
			$this->listCacheFile  = new SplFileInfo(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('local_database_path') . 'list-' . FrameworkBootstrap::getDetectedApplicationName() . '.cache');
			$this->classCacheFile = new SplFileInfo(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('local_database_path') . 'class-' . FrameworkBootstrap::getDetectedApplicationName() . '.cache');
		}

		// Set suffix and prefix from configuration
		$this->suffix = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('class_suffix');
		$this->prefix = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('class_prefix');

		// Set own instance
		self::$selfInstance = $this;

		// Skip here if no dev-mode
		if (self::$developerModeEnabled) {
			return;
		}

		// Is the cache there?
		if (FrameworkBootstrap::isReadableFile($this->listCacheFile)) {
			// Load and convert it
			$this->pendingFiles = json_decode(file_get_contents($this->listCacheFile->getPathname()));

			// List has been restored from cache!
			$this->listCached = true;
		}

		// Does the class cache exist?
		if (FrameworkBootstrap::isReadableFile($this->classCacheFile)) {
			// Then include it
			FrameworkBootstrap::loadInclude($this->classCacheFile);

			// Mark the class cache as loaded
			$this->classesCached = true;
		}
	}

	/**
	 * Tries to find the given class in our list. It will ignore already loaded 
	 * (to the program available) class/interface/trait files. But you SHOULD
	 * save below "expensive" code by pre-checking this->loadedClasses[], if
	 * possible.
	 *
	 * @param	$className	The class that shall be loaded
	 * @return	void
	 */
	private function loadClassFile (string $className) {
		// The class name should contain at least 2 back-slashes, so split at them
		//* NOISY-DEBUG: */ printf('[%s:%d] className=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $className);
		$classNameParts = explode("\\", $className);

		// Get last element
		$shortClassName = array_pop($classNameParts);

		// Create a name with prefix and suffix
		$fileName = sprintf('%s%s%s', $this->prefix, $shortClassName, $this->suffix);

		// Now look it up in our index
		//* NOISY-DEBUG: */ printf('[%s:%d] ISSET: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
		if ((isset($this->pendingFiles[$fileName])) && (!isset($this->loadedClasses[$this->pendingFiles[$fileName]->getPathname()]))) {
			// File is found and not loaded so load it only once
			//* NOISY-DEBUG: */ printf('[%s:%d] LOAD: fileName=%s - START' . PHP_EOL, __METHOD__, __LINE__, $fileName);
			FrameworkBootstrap::loadInclude($this->pendingFiles[$fileName]);
			//* NOISY-DEBUG: */ printf('[%s:%d] LOAD: fileName=%s - END' . PHP_EOL, __METHOD__, __LINE__, $fileName);

			// Count this loaded class/interface/exception
			$this->total++;

			// Mark this class as loaded for other purposes than loading it.
			$this->loadedClasses[$this->pendingFiles[$fileName]->getPathname()] = true;

			// Remove it from classes list so it won't be found twice.
			//* NOISY-DEBUG: */ printf('[%s:%d] UNSET: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
			unset($this->pendingFiles[$fileName]);

			// Developer mode excludes caching (better debugging)
			if (!self::$developerModeEnabled) {
				// Reset cache
				//* NOISY-DEBUG: */ printf('[%s:%d] classesCached=false' . PHP_EOL, __METHOD__, __LINE__);
				$this->classesCached = false;
			}
		} else {
			// Not found
			//* NOISY-DEBUG: */ printf('[%s:%d] 404: fileName=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName);
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d] EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

}
