<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\Pointer\Text;

// Import framework stuff
use Org\Mxchange\CoreFramework\Filesystem\BaseFileIo;
use Org\Mxchange\CoreFramework\Filesystem\FileIoException;
use Org\Mxchange\CoreFramework\Filesystem\Pointer\OutputPointer;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;
use \SplFileObject;

/**
 * A class for writing files
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkTextFileOutputPointer extends BaseFileIo implements OutputPointer {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Create a file pointer based on the given file. The file will also
	 * be verified here.
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @param	$mode		The output mode ('w', 'a' are valid)
	 * @throws	InvalidArgumentException	If mode is empty
	 * @throws	FileIoException			If fopen() returns not a file resource
	 * @return	void
	 */
	public static final function createFrameworkTextFileOutputPointer (SplFileInfo $fileInstance, string $mode) {
		// Some pre-sanity checks...
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEXT-FILE-OUTPUT-POINTER: fileInstance[%s]=%s,mode=%s - CALLED!', get_class($fileInstance), $fileInstance->__toString(), $mode));
		if (empty($mode)) {
			// No filename given
			throw new InvalidArgumentException('Parameter "mode" is empty');
		}

		// Try to open a handler
		$fileObject = $fileInstance->openFile($mode);

		// Is it valid?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEXT-FILE-OUTPUT-POINTER: fileObject[]=%s', gettype($fileObject)));
		if (!($fileObject instanceof SplFileObject)) {
			// Something bad happend
			throw new FileIoException($fileInstance, self::EXCEPTION_FILE_POINTER_INVALID);
		}

		// Create new instance
		$pointerInstance = new FrameworkTextFileOutputPointer();

		// Set file object
		$pointerInstance->setFileObject($fileObject);

		// Return the instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEXT-FILE-OUTPUT-POINTER: pointerInstance=%s - EXIT!', $pointerInstance->__toString()));
		return $pointerInstance;
	}

	/**
	 * Write data to a file pointer
	 *
	 * @param	$dataStream		The data stream we shall write to the file
	 * @return	mixed			Number of writes bytes or false on error
	 * @throws	NullPointerException	If the file pointer instance is not set by setFileObject()
	 * @throws	LogicException	If there is no object being set
	 */
	public function writeToFile (string $dataStream) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('TEXT-FILE-OUTPUT-POINTER: dataStream=%s - CALLED!', $dataStream));
		if (empty($dataStream)) {
			// Invalid parameter
			throw new InvalidArgumentException('Parameter "dataStream" is empty');
		} elseif (is_null($this->getFileObject())) {
			// Pointer not initialized
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		} elseif (!is_object($this->getFileObject())) {
			// Pointer is not a valid object!
			throw new LogicException(sprintf('this->fileObject[]=%s is no object', gettype($this->getFileObject())));
		}

		// Write data to the file pointer and return written bytes
		return $this->getFileObject()->fwrite($dataStream);
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function analyzeFileStructure () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes at given position by seeking to it.
	 *
	 * @param	$seekPosition	Seek position in file
	 * @param	$data			Data to be written
	 * @return	mixed			Number of writes bytes or false on error
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function writeAtPosition (int $seedPosition, string $data) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Advances to next "block" of bytes
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function next () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function valid () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Gets current seek position ("key").
	 *
	 * @return	$key	Current key in iteration
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function key () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

}
