<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Response;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Image\BaseImage;
use Org\Mxchange\CoreFramework\Manager\ManageableApplication;
use Org\Mxchange\CoreFramework\Registry\GenericRegistry;
use Org\Mxchange\CoreFramework\Response\Responseable;

/**
 * A class for an image response on an HTTP request
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * The extended headers are taken from phpMyAdmin setup tool, written by
 * Michal Cihar <michal@cihar.com>, licensed under GNU GPL 2.0.
 */
class ImageResponse extends BaseResponse implements Responseable {
	/**
	 * Instance of the image
	 */
	private $imageInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);

		// Set response type
		$this->setResponseType('image');
	}

	/**
	 * Creates an object of this class
	 *
	 * @return	$responseInstance	A prepared instance of this class
	 */
	public static final function createImageResponse () {
		// Get a new instance
		$responseInstance = new ImageResponse();

		// Return the prepared instance
		return $responseInstance;
	}

	/**
	 * Setter for image instance
	 *
	 * @param	$imageInstance	An instance of an image
	 * @return	void
	 */
	public final function setImageInstance (BaseImage $imageInstance) {
		$this->imageInstance = $imageInstance;
	}

	/**
	 * Getter for image instance
	 *
	 * @return	$imageInstance	An instance of an image
	 */
	public final function getImageInstance () {
		return $this->imageInstance;
	}

	/**
	 * Adds a cookie to the response
	 *
	 * @param	$cookieName		Cookie's name
	 * @param	$cookieValue	Value to store in the cookie
	 * @param	$encrypted		Do some extra encryption on the value
	 * @param	$expires		Timestamp of expiration (default: configured)
	 * @return	void
	 * @throws	ResponseHeadersAlreadySentException		If headers are already sent
	 * @todo	Encryption of cookie data not yet supported.
	 * @todo	If the return statement is removed and setcookie() commented out,
	 * @todo	this will send only one cookie out, the first one.
	 */
	public function addCookie (string $cookieName, $cookieValue, bool $encrypted = FALSE, int $expires = NULL) {
		// Are headers already sent?
		if (headers_sent()) {
			// Throw an exception here
			throw new ResponseHeadersAlreadySentException($this, self::EXCEPTION_HEADERS_ALREADY_SENT);
		}

		// Shall we encrypt the cookie?
		if ($encrypted) {
			// Unsupported at the moment
			$this->partialStub('Encryption is unsupported at the moment.');
		}

		// For slow browsers set the cookie array element first
		$_COOKIE[$cookieName] = $cookieValue;

		// Get all config entries
		if (is_null($expires)) {
			$expires = (time() + FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('cookie_expire'));
		}

		$path = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('cookie_path');
		$domain = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('cookie_domain');

		setcookie($cookieName, $cookieValue, $expires);
		//, $path, $domain, (isset($_SERVER['HTTPS']))
		return;

		// Now construct the full header
		$cookieString = $cookieName . '=' . $cookieValue . '; ';
		$cookieString .= 'expires=' . date('D, d-F-Y H:i:s', $expires) . ' GMT';
		// TODO Why is this not always working? $cookieString .= '; path=' . $path . '; domain=' . $domain;

		// Set the cookie as a header
		$this->cookies[$cookieName] = $cookieString;
	}

	/**
	 * Redirect to a configured URL. The URL can be absolute or relative. In
	 * case of relative URL it will be extended automatically.
	 *
	 * @param	$configEntry	The configuration entry which holds our URL
	 * @return	void
	 * @throws	ResponseHeadersAlreadySentException		If headers are already sent
	 */
	public function redirectToConfiguredUrl (string $configEntry) {
		// Get application instance
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Is the header not yet sent?
		if (headers_sent()) {
			// Throw an exception here
			throw new ResponseHeadersAlreadySentException($this, self::EXCEPTION_HEADERS_ALREADY_SENT);
		}

		// Assign application data
		$this->getTemplateInstance()->assignApplicationData($applicationInstance);

		// Get the url from config
		$url = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($configEntry . '_url');

		// Compile the URL
		$url = $this->getTemplateInstance()->compileRawCode($url);

		// Do we have a 'http' in front of the URL?
		if (substr(strtolower($url), 0, 4) != 'http') {
			// Is there a / in front of the relative URL?
			if (substr($url, 0, 1) == '/') $url = substr($url, 1);

			// No, then extend it with our base URL
			$url = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('base_url') . '/' . $url;
		}

		// Add redirect header
		$this->addHeader('Location', str_replace('&amp;', '&', $url));

		// Set correct response status
		$this->setResponseStatus('301 Moved Permanently');

		// Clear the body
		$this->setResponseBody('');

		// Flush the result
		$this->flushBuffer();

		// All done here...
		exit();
	}

	/**
	 * Flushs the cached HTTP response to the outer world
	 *
	 * @param	$force	Whether we shall force the output or abort if headers are
	 *					already sent with an exception
	 * @return	void
	 */
	public function flushBuffer (bool $force = false) {
		// Finish the image
		$this->getImageInstance()->finishImage();

		// Get image content
		$content = $this->getImageInstance()->getContent();

		// Set it as response body
		$this->setResponseBody($content);

		// Set content type
		$this->addHeader('Content-type', 'image/' . $this->getImageInstance()->getImageType());

		// Call parent method
		parent::flushBuffer($force);
	}

	/**
	 * Expires the given cookie if it is set
	 *
	 * @param	$cookieName		Cookie to expire
	 * @return	void
	 */
	public function expireCookie (string $cookieName) {
		// Is the cookie there?
		if (isset($_COOKIE[$cookieName])) {
			// Then expire it with 20 minutes past
			$this->addCookie($cookieName, '', false, (time() - 1200));

			// Remove it from array
			unset($_COOKIE[$cookieName]);
		}
	}

	/**
	 * Refreshs a given cookie. This will make the cookie live longer
	 *
	 * @param	$cookieName		Cookie to refresh
	 * @return	void
	 */
	public function refreshCookie (string $cookieName) {
		// Only update existing cookies
		if (isset($_COOKIE[$cookieName])) {
			// Update the cookie
			$this->addCookie($cookieName, $_COOKIE[$cookieName], false);
		}
	}

}
